/*
	Sam McKennoch
	skennoch at u.washington.edu
	TNN.java
	Demonstrates theta neurons
	Copyright (C) 2007 Sam McKennoch

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.*;
import java.awt.*;
import java.applet.Applet;
import java.awt.event.*;
import java.lang.Float;

class GraphPanel extends Panel
    implements Runnable, MouseListener, MouseMotionListener {

	double Io=-0.05;
	double fixed_point=0;
	double phase_ini=fixed_point+0.1;
	double phase_rad = phase_ini;
	double last_phase = phase_rad;

	int potLength = 500;
	double potential[] = new double[potLength];
	int head=0;
	int tail=potLength-1;
	int xlen = 550;
	int ylen = 300;
	int scale = 25;
	int index = 0;
    
	int xcenter = 80, ycenter = 55;
	float cx, cy; 
	int diameter=100;
	int radius=diameter/2;
	double eps=0.02;
	double dth = 0;

	int tx=xcenter+(int)diameter;
	int ty=ycenter-(int)(radius)-10;
	int sx = 25;
	int sy = ycenter + (int)(0.75*(float)diameter);					
	
	String laststring1 = "Hi";
	String laststring2 = "Hi";
	TNN graph;
 
	private volatile Thread relaxer;

	Dimension offDimension;
	Image offImage;
	Graphics offGraphics;

	GraphPanel(TNN graph) {
		this.graph = graph;
		addMouseListener(this);
	}

	//Update a frame of animation, using double-buffering to eliminate flicker
	public void update(Graphics g) {
		Dimension d = getSize();

		// Create the offscreen graphics context
		if ((offGraphics == null)
		 || (d.width != offDimension.width)
		 || (d.height != offDimension.height)) {
			offDimension = d;
			offImage = createImage(d.width, d.height);
			offGraphics = offImage.getGraphics();
		}

		// Erase the previous image
		int  cl= 238;
		Color c = new Color(cl, cl, cl);
		setBackground(c);
		offGraphics.setColor(getBackground());
		offGraphics.fillRect(0, 0, d.width, d.height);
		offGraphics.setColor(Color.black);

		// Paint the frame into the image
		paintFrame(offGraphics);

		// Paint the image onto the screen
		g.drawImage(offImage, 0, 0, null);
    }

	// Paint the previous frame (if any).
	public void paint(Graphics g) {
		if (offImage != null) {
			g.drawImage(offImage, 0, 0, null);
		}
    }

	// Paint the current frame to an image
	public void paintFrame(Graphics g) {
		//Erase the screen
		g.setColor(getBackground());
		g.fillRect(0,0,xlen,ylen);

		//Erase last phase circle
		cx = radius*(float)Math.cos(phase_rad) + xcenter - radius/10;
		cy = -radius*(float)Math.sin(phase_rad) + ycenter - radius/10;
		g.setColor(getBackground());
		g.fillArc((int)cx, (int)cy, (diameter)/10, (diameter)/10, 0, 360);

		//Update phase
		last_phase=phase_rad;
		dth=0.01*((1-Math.cos(phase_rad))+Io*(1+Math.cos(phase_rad)));
		phase_rad+=dth;
		if (phase_rad<0)
		{
			phase_rad+=2*Math.PI;
		}
		else if (phase_rad>2*Math.PI)
		{
			phase_rad-=2*Math.PI;
		}

		//Erase and redraw info strings
		ty=25;
		g.drawString(laststring1, tx, ty);
		if (Io<0)
		{
			laststring1="Fixed Point = " + ((double)Math.round(1000*(2*Math.PI-fixed_point)))/1000 + " Phase (radians) = " + ((double)Math.round(1000*phase_rad))/1000;
		}
		else
		{
			laststring1="Phase (radians) = " + ((double)Math.round(1000*phase_rad))/1000;
		}
		g.setColor(Color.black);
		g.drawString(laststring1, tx, ty);
		ty+=15;
		g.drawString(laststring2, tx, ty);
		laststring2="Scaled Potential = " + ((double)Math.round(10000*Math.tan(phase_rad/2))/10000);
		g.setColor(Color.black);
		g.drawString(laststring2, tx, ty);

		//Draw phase circle
		g.setColor(Color.black);
		g.drawArc(xcenter-radius, ycenter-radius, diameter, diameter, 0, 360);

		//Draw fixed points
		if (Io<0)
		{
			fixed_point=Math.acos((1+Io)/(1-Io));
			phase_ini = fixed_point+0.1;	
			g.setColor(Color.red);
			cx = radius*(float)Math.cos(fixed_point) + xcenter - radius/10;
			cy = -radius*(float)Math.sin(fixed_point) + ycenter - radius/10;
			g.fillArc((int)cx, (int)cy, (diameter)/10, (diameter)/10, 0, 360);
			cy = radius*(float)Math.sin(fixed_point) + ycenter - radius/10;
			g.fillArc((int)cx, (int)cy, (diameter)/10, (diameter)/10, 0, 360);
		}

		//Draw lines and labels for zero and pi
		g.setColor(Color.red);		
		g.drawLine(xcenter+(int)(radius-0.1*radius),ycenter,xcenter+(int)(radius+0.1*radius),ycenter);
		g.drawLine(xcenter-(int)(radius-0.1*radius),ycenter,xcenter-(int)(radius+0.1*radius),ycenter);
		g.drawString("0", xcenter+(int)(radius+0.15*radius), ycenter);
		g.drawString("\u03C0", xcenter-(int)(radius+0.25*radius), ycenter);

		//Draw current phase
		cx = radius*(float)Math.cos(phase_rad) + xcenter - radius/10;
		cy = -radius*(float)Math.sin(phase_rad) + ycenter - radius/10;
		g.setColor(Color.blue);
		g.fillArc((int)cx, (int)cy, (diameter)/10, (diameter)/10, 0, 360);

		//Erase Previous Potential Plot
		g.setColor(getBackground());
		drawPlot(g);

		//Draw Plot Rectangle and Info
		g.setColor(Color.black);
		g.drawRect(sx, sy,potLength, (int)diameter);
		g.drawLine(sx, sy+radius,sx+potLength, sy+radius);
		g.drawString("Potential", sx, sy+diameter+15);
		String temp = "" + -Math.round(100*((double)radius/(double)scale))/100;
		g.drawString(temp,sx-20,sy+diameter);
		temp = "" + Math.round(100*((double)radius/(double)scale))/100;
		g.drawString(temp,sx-16,sy+10);

		//Push new value onto potential stack
		potential[tail]=Math.tan(phase_rad/2);
		tail--;
		head--;
		if (tail<0)
		{
			tail+=potLength;
		}
		if (head<0)
		{
			head+=potLength;
		}

		g.setColor(Color.blue);
		drawPlot(g);
	}

	//Draw potential plot
	private void drawPlot(Graphics g) {
		index=head;
		for (int i=0; i<(potLength-1); i++)
		{
			if (index==(potLength-1))
			{
				//add this back in, it was buggy before
				//g.drawLine(potLength+xcenter+(int)diameter-i,ycenter-(int)Math.max(Math.min((scale*potential[index]),(double)radius),-(double)radius),potLength+xcenter+(int)diameter,ycenter-(int)Math.max(Math.min((scale*potential[0]),(double)radius),-(double)radius));
			}
			else {
				g.drawLine(potLength+sx-i,sy+radius-(int)Math.max(Math.min((scale*potential[index]),(double)radius),-(double)radius),potLength+sx-(i+1),sy+radius-(int)Math.max(Math.min((scale*potential[index+1]),(double)radius),-(double)radius));
			}
			index++;
			if (index==potLength)
			{
				index-=potLength;
			}
		}
	}

	public void run() {
		Thread me = Thread.currentThread();
		while (relaxer == me) {
			try {
			Thread.sleep(15);
			} catch (InterruptedException e) {
			break;
			}
			repaint();
		}
	}

	//1.1 event handling
	public void mouseClicked(MouseEvent e) {
	}

	public void mousePressed(MouseEvent e) {
		addMouseMotionListener(this);
		double bestdist = Double.MAX_VALUE;
		int x = e.getX();
		int y = e.getY();
		repaint();
		e.consume();
	}

	public void mouseReleased(MouseEvent e) {
		removeMouseMotionListener(this);
		repaint();
		e.consume();
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mouseDragged(MouseEvent e) {
		//	pick.x = e.getX();
		//	pick.y = e.getY();
		repaint();
		e.consume();
	}

	public void mouseMoved(MouseEvent e) {
	}

	public void start() {
		relaxer = new Thread(this);
		relaxer.start();
	}

	public void stop() {
		relaxer = null;
	}

}


public class TNN extends Applet implements ActionListener, ItemListener {

//Features to add:
// Show input and output spikes explicitly on the axis
// Expand to multi-layer
// Add info from clicking on the axis

	GraphPanel panel;
	Panel controlPanel;

	Button reset = new Button("Reset Io to:");
	TextField Io_box = new TextField("-0.05");

	Label spike_label = new Label("Input Spikes:");
	Button s1 = new Button("With w1:");
	TextField w1_box = new TextField("0.05");
	Button s2 = new Button("With w2:");
	TextField w2_box = new TextField("-0.2");

	public void init() {
		setLayout(new BorderLayout());

		panel = new GraphPanel(this);
		
		int  cl= 238;
		Color c = new Color(cl, cl, cl);
		panel.setBackground(c);	
		
		add("Center", panel);
		controlPanel = new Panel();
		add("South", controlPanel);

		controlPanel.setBackground(c);
		controlPanel.setLayout(new FlowLayout());
		controlPanel.setSize(550,100);

		controlPanel.add(reset); reset.addActionListener(this);
		controlPanel.add(Io_box);

		controlPanel.add(spike_label);
		controlPanel.add(s1); s1.addActionListener(this);
		controlPanel.add(w1_box);
		controlPanel.add(s2); s2.addActionListener(this);
		controlPanel.add(w2_box);

		for (int i = 0; i < panel.potLength; i++) {
			panel.potential[i]=0;
		}

		if (panel.Io<0)
		{
			panel.fixed_point=Math.acos((1+panel.Io)/(1-panel.Io));
			panel.phase_ini=panel.fixed_point+0.1;
			panel.phase_rad = panel.phase_ini;
			panel.last_phase = panel.phase_rad;		
		}
		resize(panel.xlen,panel.ylen);   
	}

	public void destroy() {
		remove(panel);
		remove(controlPanel);
	}

	public void start() {
		panel.start();
	}

	public void stop() {
		panel.stop();
	}

	public void actionPerformed(ActionEvent e) {
		Object src = e.getSource();

		if (src == reset) {
			panel.Io = Float.valueOf(Io_box.getText()).floatValue();		
			if (panel.Io<0)
			{
				panel.fixed_point=Math.acos((1+panel.Io)/(1-panel.Io));	
			}
			else
			{
				panel.fixed_point=0;
			}	
			panel.phase_ini = panel.fixed_point + 0.1;
			panel.phase_rad = panel.phase_ini;
			return;
		}
		else if (src == s1) {
				panel.phase_rad += Float.valueOf(w1_box.getText()).floatValue()*(1+Math.cos(panel.phase_rad));		
		}
		else if (src == s2) {
				panel.phase_rad += Float.valueOf(w2_box.getText()).floatValue()*(1+Math.cos(panel.phase_rad));		
		}
	}

	public void itemStateChanged(ItemEvent e) {
	}

	public String getAppletInfo() {
	return "Title: Basic Theta Neuron \nAuthor: Sam McKennoch";
	}

}